# frozen_string_literal: true

require_relative "cpk/author"
require_relative "cpk/book"
require_relative "cpk/order"
require_relative "cpk/review"
require_relative "cpk/order_agreement"
require_relative "cpk/order_tag"
require_relative "cpk/tag"
require_relative "cpk/chapter"
require_relative "cpk/post"
require_relative "cpk/posts_tag"
require_relative "cpk/comment"
require_relative "cpk/car"
require_relative "cpk/car_review"
