<% content_for :page_title do %>
Ruby on Rails Guides
<% end %>

<h1>Table of Contents</h1>
<div id="toc">
  <ul><li><a href="welcome.html">Welcome</a></li></ul>

  <% documents_by_section.each_with_index do |section, i| %>
    <h3><%= "#{i + 1}." %> <%= section['name'] %></h3>
    <ul>
      <% section['documents'].each do |document| %>
        <li>
          <a href="<%= document['url'] %>"><%= document['name'] %></a>
          <% if document['work_in_progress']%>(WIP)<% end %>
        </li>
      <% end %>
    </ul>
  <% end %>
<hr />
<ul>
  <li><a href="copyright.html">Copyright &amp; License</a></li>
</ul>
</div>
