<!DOCTYPE html>
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8"></meta>
<link rel="stylesheet" type="text/css" href="<%= digest_path('stylesheets/epub.css') %>"></link>
<title><%= yield(:page_title) || 'Ruby on Rails Guides' %></title>
</head>
<body class="guide">

  <% if content_for? :header_section %>
    <%= yield :header_section %>
    <div class="pagebreak"></div>
  <% end %>

  <% if content_for? :index_section %>
    <%= yield :index_section %>
    <div class="pagebreak"></div>
  <% end %>

  <%= yield.html_safe %>
</body>
</html>
